﻿/*
//  Распределение строк матрицы по процессам.
//
//  API
//  int MPI_GaussSeidelDataDistribution(double* pMatrix, double* pProcRows,
//                                      int* RowNum, const int ProcRank,
//                                      const int ProcNum, const int Size);
//  Input:
//      + pMatrix   - указатель типа double, исходная матрица
//                    значений функции
//      + pProcRows - указатель типа double, память под строки матрицы,
//                    посылаемые процессу
//      + RowNum    - указатель типа int
//      + ProcRank  - переменная типа int, номер процесса
//      + ProcNum   - переменная типа int, число процессов
//      + Size      - переменная типа int, размерность матрицы
//  Output:
//      + pProcRows - указатель типа double, строки матрицы,
//                    посланные процессу
//      + RowNum    - указатель типа int, число строк матрицы,
//                    посланных процессу
//  Result:
//      + функция возвращает переменную типа int, код завершения функции.
*/
int MPI_GaussSeidelDataDistribution(double* pMatrix, double* pProcRows,
                                    int* RowNum, const int ProcRank,
                                    const int ProcNum, const int Size)
{
    int Res = PARLIB_SUCCESS;

    int *pSendNum; // Число строк, отсылаемых процессу
    int *pSendInd; // Индекс первого элемента, отправляемого процессу
    int RestRows = Size;
    int i;

    assert(pProcRows);   
    assert(Size>0);     
    assert(ProcRank>-1);
    assert(ProcNum>0);

    // Выделение памяти
    pSendInd = (int*) malloc (ProcNum * sizeof(int));
    if(NULL == pSendInd)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }
    pSendNum = (int*) malloc (ProcNum * sizeof(int));
    if(NULL == pSendNum)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }

    // Определение расположения строк матрицы, отсылаемых процессу  
    *RowNum = (Size - 2) / ProcNum + 2;
    pSendNum[0] = (*RowNum)*Size;
    pSendInd[0] = 0;
    for ( i = 1; i < ProcNum; ++i ) 
    {
        RestRows = RestRows - (*RowNum) + 2;
        (*RowNum) = (RestRows - 2)/(ProcNum-i)+2;
        pSendNum[i] = (*RowNum)*Size; 
        pSendInd[i] = pSendInd[i-1]+pSendNum[i-1] - Size*2;    
    }

    // Рассылка строк  
    MPI_Scatterv(pMatrix, pSendNum, pSendInd, MPI_DOUBLE, 
        pProcRows, pSendNum[ProcRank], MPI_DOUBLE, 0, MPI_COMM_WORLD);   

    // Освобождение памяти
    free(pSendInd);
    free(pSendNum);    

    return Res;
}

/*
// Реализация итерации алгоритма Гаусса-Зейделя.
//
//  API
//  double IterationCalculation(double* pProcRows, const int Size,
//                              const int RowNum);
//
//  Input:
//      + pProcRows - указатель типа double, 
//                    строки матрицы, посланные процессу
//      + Size      - переменная типа int, размерность матрицы
//      + RowNum    - переменная типа int, число строк матрицы,
//                    посланных процессу
//  Result:
//      + функция возвращает значение типа double, максимальное отклонение,
//        полученное на этом процессе.
*/
double IterationCalculation(double* pProcRows, const int Size,
                            const int RowNum)
{
    int Res = PARLIB_SUCCESS;
    int i, j; 
    double dm, dmax,temp;  
    dmax = 0;

    // Пересчет значений элементов матрицы u
    for ( i = 1; i < RowNum - 1; ++i )
    {
        for( j = 1; j < Size - 1; ++j ) 
        {
            temp = pProcRows[Size * i + j];
            pProcRows[Size * i + j] = 0.25 * (pProcRows[Size * i + j + 1] + 
                pProcRows[Size * i + j - 1] + 
                pProcRows[Size * (i + 1) + j] + 
                pProcRows[Size * (i - 1) + j]);
            dm = fabs(pProcRows[Size * i + j] - temp);
            if (dmax < dm)
            {
                dmax = dm;
            }
        } 
    }

    return dmax;
}

/*
//  Обмен граничными данными.
//
//  API
//  int ExchangeData(double* pProcRows, const int Size, 
//                   const int RowNum, const int ProcRank,
//                   const int ProcNum);
//  Обмен происходит граничными строками
//  полос строк матрицы, посланных процессам.
//
//  Input:
//      + pProcRows - указатель типа double, строки матрицы,
//                    посланные процессу
//      + Size      - переменная типа int, размерность матрицы
//      + RowNum    - переменная типа int, число строк матрицы,
//                    посланных процессу
//      + ProcRank  - переменная типа int, номер процесса
//      + ProcNum   - переменная типа int, число процессов
//  Result:
//      + функция возвращает значение типа int, код завершения функции. 
*/
int ExchangeData(double* pProcRows, const int Size, 
                 const int RowNum, const int ProcRank,
                 const int ProcNum)
{
    int Res = PARLIB_SUCCESS;
    int NextProcNum, PrevProcNum;
    MPI_Status status;
    
    // Обмен данными между процессами 
    NextProcNum = (ProcRank == (ProcNum-1))? MPI_PROC_NULL : ProcRank + 1;
    PrevProcNum    = (ProcRank == 0)? MPI_PROC_NULL : ProcRank - 1;

    // Отправка следующему процессу и прием от предыдущего
    MPI_Sendrecv(pProcRows + Size * (RowNum - 2),Size, MPI_DOUBLE,
        NextProcNum, 4, pProcRows, Size, MPI_DOUBLE, PrevProcNum, 4,
        MPI_COMM_WORLD, &status);

    // Отправка предыдущему процессу и прием данных от следующего 
    MPI_Sendrecv(pProcRows + Size, Size, MPI_DOUBLE, PrevProcNum, 5,
        pProcRows + (RowNum - 1) * Size, Size,MPI_DOUBLE, NextProcNum, 5,
        MPI_COMM_WORLD, &status); 

    return Res;
}

/*
//  Реализация параллельного алгоритма численного решения задачи Дирихле 
//  для уравнения Пуассона методом Гаусса-Зейделя.
//
//  API
//  int MPI_GaussSeidelResultCalculation(double *pProcRows, const int Size, 
//                                       const int RowNum, const double Eps,
//                                       int* Iterations, const int ProcRank,
//                                       const int ProcNum);
//
//  Используя пятиточечный шаблон для вычисления значений производных,
//  мы можем представить уравнение Пуассона в конечно-разностной форме:
//  u[i][j] = 1/4*(u[i-1][j] + u[i+1][j] + u[i][j-1] + u[i][j+1]).
//  Это служит основой для построения различных
//  итерационных схем решения задачи Дирихле.
//  Метод Гаусса-Зейделя для проведения итераций уточнения
//  использует правило:
//  u[i][j]^k = 1/4*(u[i-1][j]^k + u[i+1][j]^(k-1) 
//                  + u[i][j-1]^k + u[i][j+1]^(k-1)),
//  по которому очередное k-ое приближение значения u[i][j],
//  вычисляется по последнему k-ому приближению значений u[i-1][j] и u[i][j-1],
//  и предпоследнему (k-1)-ому приближению значений u[i+1][j] и u[i][j+1].
//  Итерации метода повторяются до достижения требуемой точности.
//
//  Пересчет значений исходной матрицы функции u происходит отдельно,
//  каждая полоса строк на своем процессе. На каждой итерации процессы
//  обмениваются граничными строками.
//
//  Input:
//      + pProcRows  - указатель типа double, строки матрицы,
//                     посланные процессу
//      + Size       - переменная типа int, размерность матрицы
//      + RowNum     - переменная типа int, число строк матрицы,
//                     посланных процессу
//      + Eps        - переменная типа double, требуемая точность вычислений
//      + Iterations - указатель типа int
//      + ProcRank   - переменная типа int, номер процесса
//      + ProcNum    - переменная типа int, число процессов
//  Output:
//      + Iterations - указатель типа int, число выполненных итераций алгоритма
//  Result:
//      + функция возвращает значение типа int, код завершения функции. 
*/
int MPI_GaussSeidelResultCalculation(double *pProcRows, const int Size, 
                                     const int RowNum, const double Eps,
                                     int* Iterations, const int ProcRank,
                                     const int ProcNum)
{
    int Res = PARLIB_SUCCESS;

    double ProcDelta, Delta;

    assert(pProcRows);   
    assert(ProcNum == 2);
    assert(Size>0); 
    assert(Eps>0);
    assert(RowNum>0);
    assert(ProcRank>-1);
    assert(ProcNum>0);

    *Iterations = 0;
    do 
    {
        (*Iterations)++;
        // Обмен граничными строками
        // полосы строк процесса
        ExchangeData(pProcRows, Size, RowNum, ProcRank, ProcNum);

        // Итерация метода Гаусса-Зейделя
        ProcDelta = IterationCalculation(pProcRows, Size, RowNum);

        // Вычисление максимального значения отклонения
        MPI_Allreduce(&ProcDelta, &Delta, 1,MPI_DOUBLE, MPI_MAX,
            MPI_COMM_WORLD);

    } while ( Delta > Eps);

    return Res;
}
