﻿/*
//  Реализация алгоритма численного решения задачи Дирихле 
//  для уравнения Пуассона методом Гаусса-Зейделя с использованием
//  технологии OpenMP.
//
//  API
//  int OMP_BaseGaussSeidelCalculation(double** u, double** f, 
//                                     const double eps, const int Size);
//
//  Используя пятиточечный шаблон для вычисления значений производных,
//  мы можем представить уравнение Пуассона в конечно-разностной форме:
//  u[i][j] = 1/4*(u[i-1][j] + u[i+1][j] + u[i][j-1] + u[i][j+1] - h*h*f[i][j]).
//  Это служит основой для построения различных
//  итерационных схем решения задачи Дирихле.
//  Метод Гаусса-Зейделя для проведения итераций уточнения
//  использует правило:
//  u[i][j]^k = 1/4*(u[i-1][j]^k + u[i+1][j]^(k-1) 
//                  + u[i][j-1]^k + u[i][j+1]^(k-1) - h*h*f[i][j]),
//  по которому очередное k-ое приближение значения u[i][j],
//  вычисляется по последнему k-ому приближению значений u[i-1][j] и u[i][j-1],
//  и предпоследнему (k-1)-ому приближению значений u[i+1][j] и u[i][j+1].
//  Итерации метода повторяются до достижения требуемой точности.
//
//  Параллелизм организуем путем 
//  произвольного порядка пересчета значений u[i][j]. 
//
//  Input:
//      + u    - указатель на вектор памяти элементов типа double*,
//               матрица дискретного представления функции u
//      + f    - указатель на вектор памяти элементов типа double*,
//               матрица дискретного представления функции f
//               (правая часть уравнения Пуассона)
//      + eps  - переменная типа double, требуемая точность вычислений
//      + Size - переменная типа int, размерность матрицы
//  Output:
//      + u    - указатель на вектор памяти элементов типа double*,
//               матрица дискретного представления решения
//  Result:
//      + функция возвращает значение типа int, код завершения функции.
*/
int OMP_BaseGaussSeidelCalculation(double** u, double** f,
                                   const double eps, const int Size)
{
    int Res = PARLIB_SUCCESS;
    int i, j;  
    double dmax, temp, dm;   
    double h = 1.0 / (double)Size;
    omp_lock_t dmax_lock;
    omp_init_lock (&dmax_lock);

    assert(Size > 0);
    assert(eps > 0);
    assert(u);
    for( i = 0; i < Size + 2; ++i )
    {
        assert(u[i]);
    }
    assert(f);
    for( i = 0; i < Size + 2; ++i )
    {
        assert(f[i]);
    }

    do 
    {	
        dmax = 0; // Максимальное изменение значений u
#pragma omp parallel for
        for ( i = 1; i < Size + 1; ++i ) 
        {
#pragma omp parallel for private(temp, dm)
            for ( j = 1; j< Size + 1; ++j ) 
            {
                // Пересчет значений u[i][j]
                temp = u[i][j];
                u[i][j] = 0.25 * (u[i-1][j] + u[i+1][j] +
                    u[i][j-1] + u[i][j+1] - h * h * f[i][j]);
                dm = fabs(temp - u[i][j]);

                // Выделение максимального изменения u
                omp_set_lock(&dmax_lock);
                if ( dmax < dm )
                {
                    dmax = dm;
                }
                omp_unset_lock(&dmax_lock);
            } // конец вложенной параллельной области
        } // конец внешней параллельной области       
    } while ( dmax > eps );

    return Res;
}
