﻿/*
//  Реализация операции выбора
//  минимального значения A или B.
//
//  API
//  double Min(const double A, const double B);
//
//  Операция выбора минимального значения должна учитывать
//  способ указания в матрице смежности несуществующих дуг графа.
//
//  Input:
//      + A - переменная типа double, вес дуги графа
//      + B - переменная типа double, вес дуги графа  
//  Result:
//      + функция возвращает минимальное значение
//	      или минус единицу, если в матрице смежности
//		  дуг графа не существует
*/
__inline double Min(const double A, const double B)
{
    double Result = (A < B) ? A : B;

    // Если вершины нет в матрице смежности
    if((A < 0) && (B >= 0)) Result = B;
    if((B < 0) && (A >= 0)) Result = A;
    if((A < 0) && (B < 0))  Result = -1;

    return Result;
}

/*
//  Реализация последовательного алгоритма Флойда,
//  поиска минимальных расстояний 
//  между всеми парами вершин графа.
//
//  API
//  int SerialFloydCalculation(double* pMatrix, const int Size);
//  
//  В ходе выполнения алгоритма матрица смежности A изменяется,
//  после завершения вычислений в матрице A будет храниться 
//  требуемый результат - длины минимальных путей 
//  для каждой пары вершин исходного графа.
//
//  Input:
//      + pMatrix - указатель на вектор памяти элеметов типа double,
//	                матрица смежности графа
//	    + Size    - переменная типа int, размерность матрицы смежности
//  Output:
//      + pMatrix - указатель на вектор памяти элеметов типа double,
//	                изменённая матрица смежности, в которой будут храниться
//				    длины минимальных путей для каждой пары вершин исходного графа
//  Result:
//      + функция возвращает значение типа int, код заверщения функции
*/
int SerialFloydCalculation(double* pMatrix, const int Size)
{
    int Res = PARLIB_SUCCESS;
    int i, j, k;
    double t1, t2;

    assert(Size > 0);
    assert(pMatrix);

    // В ходе выполнения алгоритма матрица смежности изменяется,
    // после завершения вычислений в матрице будет храниться требуемый результат - 
    // длины минимальных путей для каждой пары вершин исходного графа
    for( k = 0; k < Size; ++k )
    {
        for( i = 0; i < Size; ++i )
        {
            for( j = 0; j < Size; ++j )
            {
                if((pMatrix[i * Size + k] != -1) &&
                    (pMatrix[k * Size + j] != -1)) 
                {
                        t1 = pMatrix[i * Size + j];
                        t2 = pMatrix[i * Size + k] + pMatrix[k * Size + j];
                        pMatrix[i * Size + j] = Min(t1, t2);
                }
            }
        }
    }

    return Res;
}
