﻿/*
//  Реализация последовательного алгоритма
//  вычисления решения системы n линейных уравнений методом Гаусса.
//
//  API
//  int SerialGaussianCalculation(double* pMatrix, double* pVector,
//                                double* pResult, const int Size);
//
//  В матричном виде система может быть представлена A*x = b,
//  где A - вещественная матрица коэффициентов при неизвестных размера n*n,
//  а вектора b и x состоят из n элементов.
//  Под задачей решения системы линейных уравнений для заданных матрицы А и вектора b
//  понимается нахождение значения вектора неизвестных x,
//  при котором выполняются все уравнения системы.
//
//  Input:
//      + pMatrix - указатель на вектор памяти элементов типа double,
//                  матрица коэффициентов при неизвестных,
//                  i-я строка начинается с Size*i индекса
//      + pVector - указатель на вектор памяти элементов типа double,
//                  правая часть системы линейных уравнений
//      + Size    - переменная типа int, размерность системы линейных уравнений
//  Output:
//      + pResult - указатель на вектор памяти элементов типа double,
//                  решение системы линейных уравнений
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int SerialGaussianCalculation(double* pMatrix, double* pVector, 
                              double* pResult, const int Size)
{

    int Res = PARLIB_SUCCESS;
    int* pPivotPos;   // Определяет номера строк матрицы, выбираемых в качестве ведущих,
                      // по итерациям прямого хода метода Гаусса.
    int* pPivotIter;  // Номера итераций прямого хода метода Гаусса,
                      // на которых строки использовались в качестве ведущих.
                      // Строка i выбиралась ведущей на итерации pPivotIter[i]
    int i;


    pPivotPos  = (int*) malloc (Size * sizeof(int));
    assert(pPivotPos); 
    pPivotIter = (int*) malloc (Size * sizeof(int));
    assert(pPivotIter);


    assert(pMatrix);
    assert(pVector);
    assert(pResult);
    assert(Size > 0);

    // -1 - строка не выбиралась ещё в качестве ведущей
    for ( i = 0; i < Size; ++i ) 
    {
        pPivotIter[i] = -1;
    }
    for( i = 0; i < Size; ++i )
    {
        pResult[i] = 0.0;
    }
    // Последовательное исключение неизвестных (прямой ход метода Гаусса)
    SerialGaussianElimination (pMatrix, pVector, pPivotPos, pPivotIter, Size);
    // Определение значений неизвестных (обратный ход метода Гаусса)
    SerialBackSubstitution (pMatrix, pVector, pResult, pPivotPos, Size);

    free(pPivotPos);
    free(pPivotIter);
    return Res;
}

/*
//  Реализация последовательного алгоритма
//  прямого хода метода Гаусса.
//
//  API
//  int SerialGaussianElimination(double* pMatrix, double* pVector,
//                                int* pPivotPos, int* pPivotIter,
//                                const int Size);
//
//  Прямой ход метода Гаусса состоит в последовательном исключении неизвестных
//  в уравнениях решаемой системы линейных уравнений. 
//  На i-ой итерации метода производится исключение неизвестной i для всех уравнений
//  с номерами k, больших i. 
//  Для этого из этих уравнений осуществляется вычитание строки i,
//  умноженной на константу (Aki/Aii) с тем, 
//  чтобы результирующий коэффициент при неизвестной Xi в строках оказался нулевым.
//
//  Input:
//      + pMatrix    - указатель на вектор памяти элементов типа double,
//                     матрица коэффициентов при неизвестных,
//                     i-я строка начинается с Size*i индекса
//      + pVector    - указатель на вектор памяти элементов типа double,
//                     правая часть системы линейных уравнений
//      + pPivotPos  - указатель на вектор памяти элементов типа int,
//                     номера строк матрицы, выбираемых в качестве ведущих,
//                     по итерациям прямого хода метода Гаусса
//      + pPivotIter - указатель на вектор памяти элементов типа int,
//                     номера итераций прямого хода метода Гаусса,
//                     на которых строки использовались в качестве ведущих
//      + Size       - переменная типа int, 
//                     размерность системы линейных уравнений
//  Output:
//      + pResult    - указатель на вектор памяти элементов типа double,
//                     решение системы линейных уравнений
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int SerialGaussianElimination(double* pMatrix, double* pVector,
                              int* pPivotPos, int* pPivotIter,
                              const int Size) 
{
    int Res = PARLIB_SUCCESS;
    int Iter;       // Число итераций
    int PivotRow;   // Номер текущей ведущей строки

    assert(pMatrix);
    assert(pVector);
    assert(Size > 0);

    for ( Iter = 0; Iter < Size; ++Iter ) 
    {
        // Найти ведущую строку
        PivotRow = SerialFindPivotRow(pMatrix, pPivotIter, Size, Iter);
        pPivotPos[Iter] = PivotRow;
        pPivotIter[PivotRow] = Iter;
        // Исключить неизвестные в столбце Iter
        SerialColumnElimination(pMatrix, pVector, pPivotIter, PivotRow, Iter, Size);
    }
    return Res;
}

/*
//  Реализация последовательного алгоритма
//  поиска ведущей строки на данной итерации.
//  
//  API
//  int SerialFindPivotRow(double* pMatrix, int* pPivotIter,
//                         const int Size, const int Iter);
//
//  В качестве ведущей строки выбирается строка, 
//  которая ранее не использовалась в качестве ведущей,
//  с максимальным по абсолютному значению элементом, 
//  расположенным в столбце Iter.  
//
//  Input:
//      + pMatrix    - указатель на вектор памяти элементов типа double,
//                     матрица коэффициентов при неизвестных,
//                     i-я строка начинается с Size*i индекса
//      + pPivotIter - указатель на вектор памяти элементов типа int,
//                     номера итераций прямого хода метода Гаусса,
//                     на которых строки использовались в качестве ведущих
//      + Size       - переменная типа int, размерность 
//                     системы линейных уравнений  
//      + Iter       - переменная типа int, номер текущего столбца
//                     для поиска максимального по модулю элемента 
//  Result:
//      + функция возвращает значение типа int, номер ведущей строки
*/
int SerialFindPivotRow(double* pMatrix, int* pPivotIter,
                       const int Size, const int Iter) 
{
    int PivotRow = -1;    // Номер ведущей строки
    double MaxValue =  0; // Значение ведущего элемента
    int i;     

    assert(pMatrix);
    assert(Iter >= 0);
    assert(Size > 0);

    // Выбрать строку с максимальным ведущим элементом
    for ( i = 0; i < Size; ++i ) 
    {
        if ((pPivotIter[i] == -1) && 
            (fabs(pMatrix[i*Size+Iter]) > MaxValue)) 
        {
                PivotRow = i;
                MaxValue = fabs(pMatrix[i * Size + Iter]);
        }
    }
    return PivotRow;
}

/*
//  Реализация последовательного алгоритма
//  вычитания ведущей строки из строк,
//  которые еще не использовались в качестве ведущих.
//  Используется при вычислении решения системы линейных уравнений
//  методом Гаусса
//
//  API
//  int SerialColumnElimination(double* pMatrix, double* pVector, int* pPivotIter,
//                              int const Pivot, const int Iter, const int Size)
//
//  Input:
//      + pMatrix    - указатель на вектор памяти элементов типа double,
//                     матрица коэффициентов,
//                     i-я строка начинается с Size*i индекса
//      + pVector    - указатель на вектор памяти элементов типа double,
//                     правая часть системы линейных уравнений
//      + pPivotIter - указатель на вектор памяти элементов типа int,
//                     номера итераций прямого хода метода Гаусса,
//                     на которых строки использовались в качестве ведущих
//      + Pivot      - переменная типа int, номер ведущей строки
//      + Iter       - переменная типа int, номер столбца, в котором надо
//                     занулить все значения, кроме элемента в ведущей строке 
//      + Size       - переменная типа int, 
//                     размерность системы линейных уравнений	              
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int SerialColumnElimination(double* pMatrix, double* pVector, int* pPivotIter,
                            int const Pivot, const int Iter, const int Size)
{
    int Res = PARLIB_SUCCESS;
    int i, j;
    double PivotValue, PivotFactor; 

    assert(pMatrix);
    assert(pVector);
    assert(Pivot >= 0);
    assert(Iter >= 0);
    assert(Size > 0);

    PivotValue = pMatrix[Pivot * Size + Iter];
    // Исключить неизвестные в столбце Iter
    for ( i = 0; i < Size; ++i ) 
    {
        if (pPivotIter[i] == -1) 
        {
            PivotFactor = pMatrix[i * Size + Iter] / PivotValue;
            for ( j = Iter; j < Size; ++j ) 
            {
                pMatrix[i * Size + j] -= PivotFactor * pMatrix[Pivot * Size + j];
            }
            pVector[i] -= PivotFactor * pVector[Pivot];
        }
    }  
    return Res;
}

/*
//  Реализация последовательного алгоритма
//  обратного хода метода Гаусса.
//
//  API
//  int SerialBackSubstitution (double* pMatrix, double* pVector, double* pResult,
//                              int* pPivotPos, const int Size);
//
//  После приведения матрицы коэффициентов к верхнему треугольному виду
//  во время прямого хода метода Гаусса,
//  становится возможным определение значений неизвестных.
//  Из последнего уравнения преобразованной системы 
//  может быть вычислено значение переменной Xn-1,
//  после этого из предпоследнего уравнения становится возможным 
//  определение переменной Xn-2 и т.д.
//
//  Input:
//      + pMatrix   - указатель на вектор памяти элементов типа double,
//                    приведенная прямым ходом Гаусса матрица коэффициентов,
//                    i-я строка начинается с Size*i индекса
//      + pVector   - указатель на вектор памяти элементов типа double,
//                    приведенная прямым ходом Гаусса правая часть
//                    системы линейных уравнений
//      + pPivotPos - указатель на вектор памяти элементов типа int,
//                    номера строк матрицы, выбираемых в качестве ведущих,
//                    по итерациям прямого хода метода Гаусса
//      + Size      - переменная типа int, размерность системы линейных уравнений
//  Output:
//      + pResult   - указатель на вектор памяти элементов типа double,
//                    решение системы линейных уравнений
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int SerialBackSubstitution (double* pMatrix, double* pVector, double* pResult,
                            int* pPivotPos, const int Size) 
{
    int Res = PARLIB_SUCCESS;
    int i, j;
    int RowIndex, Row;

    assert(pMatrix);
    assert(pVector);
    assert(pResult);
    assert(Size > 0);

    for ( i = Size - 1; i >= 0; --i ) 
    {
        RowIndex = pPivotPos[i];
        pResult[i] = pVector[RowIndex] / pMatrix[Size * RowIndex + i];
        for ( j = 0; j < i; ++j ) 
        {
            Row = pPivotPos[j];
            pVector[Row] -= pMatrix[Row*Size+i] * pResult[i];
            pMatrix[Row * Size + i] = 0;
        }
    }
    return Res;
} 
