﻿/*
//  Распределение данных по процессам для реализации
//  перемножения матрицы на вектор 
//  при ленточном разделении данных по строкам.
//
//  API
//  int MPI_RowSplitMatrVectMultDataDistr(double* pMatrix, 
//                                        double* pProcRows, 
//                                        double* pVector,
//                                        int RowNum,
//                                        const int ProcRank,
//                                        const int ProcNum,
//                                        const int Size);
//
//  Строки матрицы A распределяются между всеми процессами,
//  вектор b дублируется.
//
//  Input:
//      + pMatrix   - указатель на вектор памяти элементов типа double,
//                    матрица A
//      + pProcRows - указатель на векор памяти элементов типа double,
//                    часть строк матрицы, посылаемых процессу
//      + pVector   - адрес указателя на вектор памяти элементов типа double,
//                    вектор b
//      + RowNum    - переменная типа int, число строк матрицы,
//                    отправляемых процессу
//      + ProcRank  - переменная типа int, порядковый номер процесса 
//      + ProcNum   - переменная типа int, общее число процессов 
//      + Size      - переменная типа int, размерность матирцы
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int MPI_RowSplitMatrVectMultDataDistr(double* pMatrix, 
                                      double* pProcRows, 
                                      double* pVector,
                                      int RowNum,
                                      const int ProcRank,
                                      const int ProcNum,
                                      const int Size)
{
    int Res = 0;
    int *pSendNum;     // Число элементов, посланных процессу
    int *pSendInd;     // Индекс первого посланного элемента процессу 
    int RestRows=Size; // Число строк, которые ещё не распределены
    int i;

    // Разослать по процессам вектор, на который умножается матрица
    MPI_Bcast(pVector, Size, MPI_DOUBLE, 0, MPI_COMM_WORLD); 
    
    // Выделение памяти
    pSendInd = (int*) malloc (ProcNum * sizeof(int)); 
    if(NULL == pSendInd)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }
    pSendNum = (int*) malloc (ProcNum * sizeof(int)); 
    if(NULL == pSendNum)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }
    
    // Определить расположение обрабатываемых строк матрицы
    // для текущего процесса
    RowNum = (Size/ProcNum); 
    pSendNum[0] = RowNum*Size; 
    pSendInd[0] = 0; 
    for ( i = 1; i < ProcNum; ++i ) 
    { 
        // Определение числа ещё нераспределённых строк
        RestRows -= RowNum; 
        // Определение числа строк для текущего процесса
        RowNum = RestRows/(ProcNum-i); 
        // Вычисление количества отсылаемых элементов текущему процессу
        pSendNum[i] = RowNum*Size; 
        // Вычисление индекса первого отсылаемого элемента
        // для текущего процесса
        pSendInd[i] = pSendInd[i-1]+pSendNum[i-1]; 
    } 

    // Рассылка каждому процессу своей порции строк матрицы
    // для перемножения на вектор
    MPI_Scatterv(pMatrix , pSendNum, pSendInd, MPI_DOUBLE, pProcRows,  
        pSendNum[ProcRank], MPI_DOUBLE, 0, MPI_COMM_WORLD); 

    // Освобождение памяти
    free(pSendNum); 
    free(pSendInd);       

    return Res; 
}

/*
//  Вычисление части вектора результата на каждом
//  процессе при ленточном разделении данных по строкам.
//
//  API
//  int MPI_RowSplitMatrVectMultResultCalc(double* pProcRows, 
//                                         double* pVector, 
//                                         double* pProcResult,  
//                                         const int RowNum, 
//                                         const int Size); 
//  Часть вектора результата вычисляется как скалярное произведение
//  соответствующей строки матрицы A на вектор b.
//
//  Input:
//      + pProcRows   - указатель на векор памяти элементов типа double,
//                      часть строк матрицы, посылаемых процессу
//      + pVector     - адрес указателя на вектор памяти элементов типа double,
//                      вектор b
//      + pProcResult - указатель на векор памяти элементов типа double,
//                      часть вектора результата, получаемая данным процессом
//      + RowNum      - переменная типа int, число строк матрицы,
//                      отправляемых процессу
//      + Size        - переменная типа int, размерность матирцы
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int MPI_RowSplitMatrVectMultResultCalc(double* pProcRows, 
                                       double* pVector, 
                                       double* pProcResult,  
                                       const int RowNum, 
                                       const int Size)
{
    int Res = PARLIB_SUCCESS;
    int i, j;

    assert(pProcRows);
    assert(pVector);
    assert(pProcResult);
    assert(Size > 0);
    assert(RowNum > -1);
    
    // Вычислить часть вектора результата
    // на каждом процессе
    for ( i = 0; i < RowNum; ++i ) 
    { 
        pProcResult[i] = 0; 
        for ( j = 0; j < Size; ++j ) 
        {
            pProcResult[i] += pProcRows[i*Size+j] * pVector[j]; 
        }
    } 

    return Res;
}

/*
//  Сбор частей вектора результата воедино
//  на всех процессах при ленточном разделении 
//  данных по строкам.
//
//  API
//  int MPI_RowSplitMatrVectMultResultRepl(double* pProcResult, 
//                                         double* pResult, 
//                                         const int ProcRank,
//                                         const int ProcNum,
//                                         const int Size);
//
//  Input:
//      + pProcResult - указатель на векор памяти элементов типа double,
//                      часть вектора результата, получаемая данным процессом
//      + pResult     - указатель на векор памяти элементов типа double,
//                      вектор результат - c
//      + ProcRank    - переменная типа int, порядковый номер процесса 
//      + ProcNum     - переменная типа int, общее число процессов 
//      + Size        - переменная типа int, размерность матирцы
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int MPI_RowSplitMatrVectMultResultRepl(double* pProcResult, 
                                       double* pResult, 
                                       const int ProcRank,
                                       const int ProcNum,
                                       const int Size)
{
    int Res = PARLIB_SUCCESS;
    int i;            
    int *pReceiveNum;  // Число элементов, отсылаемых текущим процессом
    int *pReceiveInd;  // Индекс первого отсылаемого элемента текущим процессом
                       // вектора результата
    int RestRows=Size; // Число нераспределённых строк 

    assert(pProcResult);
    assert(pResult);
    assert(ProcNum > 0);
    assert(Size > 0);
    assert(ProcRank > -1);

    // Выделение памяти
    pReceiveNum = (int*) malloc (ProcNum * sizeof(int));    
    if(NULL == pReceiveNum)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }
    pReceiveInd = (int*) malloc (ProcNum * sizeof(int));   
    if(NULL == pReceiveInd)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }

    // Определение расположения пересылаемого блока 
    // результирующего вектора текущего процесса
    pReceiveInd[0] = 0; 
    pReceiveNum[0] = Size/ProcNum; 
    for ( i = 1; i < ProcNum; ++i ) 
    { 
        RestRows -= pReceiveNum[i-1]; 
        pReceiveNum[i] = RestRows/(ProcNum-i); 
        pReceiveInd[i] = pReceiveInd[i-1]+pReceiveNum[i-1]; 
    } 

    // Сбор всего результирующего вектора на каждом процессе
    MPI_Allgatherv(pProcResult, pReceiveNum[ProcRank], MPI_DOUBLE, pResult,  
        pReceiveNum, pReceiveInd, MPI_DOUBLE, MPI_COMM_WORLD); 

    // Освобождение памяти
    free(pReceiveNum);  
    free(pReceiveInd); 

    return Res;
}