﻿/*
//  Распределение строк матрицы смежности графа по процессам.
//
//  API
//  int MPI_PrimDataDistribution(double* pMatrix, double* pProcRows,
//                               const int ProcRank, const int ProcNum,
//                               const int Size);
//
//  Строки распределяются между процессами поровну.
//
//  Input:
//      + pMatrix   - указатель на вектор элементов типа double,
//                    матрица смежности графа
//      + pProcRows - указатель на вектор элементов типа double,
//                    память под строки матрицы, отправляемые процессу
//      + ProcRank  - переменная типа int, номер процесса
//      + ProcNum   - переменная типа int, число процессов
//      + Size      - переменная типа int, размерность матирцы
//  Output:
//      + pProcRows - указатель на вектор элементов типа double,
//                    строки матрицы, отправленные процессу
//  Result:
//      + функция возвращает значение типа int, код завершение функции
*/
int MPI_PrimDataDistribution(double* pMatrix, double* pProcRows,
                             const int ProcRank, const int ProcNum,
                             const int Size)
{
    int Res = PARLIB_SUCCESS;
    int RowNum = Size / ProcNum;

    assert(NULL != pProcRows);
    assert(ProcRank > -1);
    assert(ProcNum > 0);
    assert(Size > 0);

    MPI_Scatter(pMatrix, RowNum * Size, MPI_DOUBLE, 
                pProcRows, RowNum * Size, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    return Res;
}

/*
//  Реализация алгоритма Прима, 
//  построения минимального охватывающего дерева.
//
//  API
//  int MPI_PrimResultCalculation(TTreeNode** pSceleton, double* pProcRows,
//                                const double MAX_VALUE, const int ProcRank,
//                                const int ProcNum, const int Size);
//
//  Алгоритм начинает работу с произвольной вершины графа,
//  выбираемой в качестве корня дерева, и в ходе последовательно 
//  выполняемых итераций расширяет конструируемое дерево до
//  минимального охватывающего дерева (МОД).
//
//  Действия, выполняемые на каждой итерации алгоритма 
//  являются независимыми и могут реализовываться одновременно.
//
//  Общая схема параллельного выполнения алгоритма Прима:
//      1. Определяется вершина t графа G, имеющая дугу 
//         минимального веса d[t] до множества Vt,
//	       для выбора такой вершины необходимо осуществить
//         поиск минимума в наборах величин d[i],
//		   имеющихся на каждом из вычислительных элементов,
//		   и выполнить редукцию полученных значений
//      2. Номер выбранной вершины для включения в охватывающее дерево
//	       передается всем вычислительным элементам
//      3. Обновляются наборы величин d[i] с учетом добавления новой вершины
//
//  Input:
//      + pProcRows - указатель на вектор памяти элеметов типа double,
//	                  строки матрицы смежности графа
//	    + pSceleton - указатель типа TTreeNode* на минимальное охватывающее дерево 
//      + MAX_VALUE - переменная типа double, максимальное значение пути
//      + ProcRank  - переменная типа int, номер процесса
//      + ProcNum   - переменная типа int, число процессов
//	    + Size      - переменная типа int, размерность матрицы смежности
//  Output:
//      + pSceleton - указатель типа TTreeNode* на полностью 
//	                  созданное минимальное охватывающее дерево
//  Result:
//      + функция возвращает значение типа int, код заверщения функции
*/
int MPI_PrimResultCalculation(TTreeNode** pSceleton, double* pProcRows,
                              const double MAX_VALUE, const int ProcRank,
                              const int ProcNum, const int Size)
{
    int Res = PARLIB_SUCCESS;

    int LastAdded;
    int NearestNode;
    TTreeNode* tmp;
    double t1, t2;
    double MinDistance;
    int Parent;
    int i, Iter;
    int RowNum = Size / ProcNum;
    double* MinDist;  // Минимальное расстояние 
    int* NearNodeNum; // Номер вершины с минимальным расстоянием
    int MinInd;       // Индекс минимального расстояния 
                      // в массиве минимальных расстояний текущей итерации
    int PivotProcess; // Номер процесса, содержащего ближайшую вершину
    int ParentNodeNum;
    int RowCount = 0;
    int TmpNodeNum;
    MPI_Status Status;    
    TGraphNode** NotInSceleton;  // Вершины графа, не помещенные в МОД

    assert(NULL != pProcRows);
    assert(ProcRank > -1);
    assert(ProcNum > 0);
    assert(MAX_VALUE > 0.0);
    assert(Size > 0);

    // Выделение памяти
    MinDist = (double*) malloc (ProcNum * sizeof(double));
    if(NULL == MinDist)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }

    NearNodeNum = (int*) malloc (ProcNum * sizeof(int));
    if(NULL == NearNodeNum)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }
    if(0 == ProcRank)
    {
        for ( i = 0; i < Size; ++i )
        {
            pSceleton[i] = NULL;
        }
    }

    // Инициализация массива информации о вершинах,
    // невключенных в состав МОД
    NotInSceleton = (TGraphNode**) malloc ((Size - 1) * sizeof(TGraphNode*));
    if(NULL == NotInSceleton)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }
    for ( i = 0; i < Size - 1; ++i )
    {
        NotInSceleton[i] = (TGraphNode*) malloc (sizeof(TGraphNode));
        NotInSceleton[i]->NodeNum = i+1;
        NotInSceleton[i]->Distance = -1.0f;
        NotInSceleton[i]->ParentNodeNum = -1;
    }

    // Начальная вершина
    LastAdded = 0;

    // Итерации алгоритма Прима
    for ( Iter = 1; Iter < Size; ++Iter )
    {
        if(ProcRank == ProcNum - 1)
        {
            // Пересчёт расстояния от вершин, не включенных в состав МОД, до МОД
            for ( i = 0; i < RowNum - 1; ++i )
            {
                if (NotInSceleton[i + ProcRank * RowNum] != NULL)
                {
                    t1 = NotInSceleton[i + ProcRank * RowNum]->Distance;
                    t2 = pProcRows[(NotInSceleton[i + ProcRank * RowNum]->NodeNum % RowNum) * Size 
                                  + LastAdded];
                    if (((t1<0) && (t2>0)) || ((t1>0) && (t2>0) && (t1>t2)))
                    {
                        NotInSceleton[i + ProcRank * RowNum]->Distance = t2;
                        NotInSceleton[i + ProcRank * RowNum]->ParentNodeNum = LastAdded;
                    }
                }
            }

            // Выбор вершины с минимальным расстоянием (с учетом -1)
            NearestNode = -1;
            MinDistance = MAX_VALUE;        
            for ( i = 0; i < RowNum - 1; ++i )
            {
                if (NotInSceleton[i + ProcRank * RowNum] != NULL)
                {
                    t1 = NotInSceleton[i + ProcRank * RowNum]->Distance;
                    t2 = MinDistance;
                    if ((t1 > 0) && (t1 < t2))
                    {
                        MinDistance = t1;
                        NearestNode = NotInSceleton[i + ProcRank * RowNum]->NodeNum;
                    } 
                }
            }
        }// if(ProcRank == ProcNum - 1)
        else
        {
            // Пересчёт расстояния от вершин, не включенных в состав МОД, до МОД
            for ( i = 0; i < RowNum; ++i )
            {
                if (NotInSceleton[i + ProcRank * RowNum] != NULL)
                {
                    t1 = NotInSceleton[i + ProcRank * RowNum]->Distance;
                    t2 = pProcRows[(NotInSceleton[i + ProcRank * RowNum]->NodeNum % RowNum) * Size 
                                  + LastAdded];
                    if (((t1<0) && (t2>0)) || ((t1>0) && (t2>0) && (t1>t2)))
                    {
                        NotInSceleton[i + ProcRank * RowNum]->Distance = t2;
                        NotInSceleton[i + ProcRank * RowNum]->ParentNodeNum = LastAdded;
                    }
                }
            }

            // Выбор вершины с минимальным расстоянием (с учетом -1)
            NearestNode = -1;
            MinDistance = MAX_VALUE;           
            for ( i = 0; i < RowNum; ++i )
            {
                if (NotInSceleton[i + ProcRank * RowNum] != NULL)
                {
                    t1 = NotInSceleton[i + ProcRank * RowNum]->Distance;
                    t2 = MinDistance;
                    if ((t1 > 0) && (t1 < t2))
                    {
                        MinDistance = t1;
                        NearestNode = NotInSceleton[i + ProcRank * RowNum]->NodeNum;
                    }
                }
            }
        }// else {}        
        
        // Инициализация массива минимальных расстояний и
        // соответствующих номеров вершин
        MPI_Gather(&MinDistance, 1, MPI_DOUBLE, 
                   MinDist, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
        MPI_Gather(&NearestNode, 1, MPI_INT, 
                   NearNodeNum, 1, MPI_INT, 0, MPI_COMM_WORLD);

        // Выбор минимального расстояния и соответствующего номера вершины
        if(0 == ProcRank)
        {
            MinInd = 0;
            for( i = 1; i < ProcNum; ++i )
            {
                if(MinDist[i] < MinDist[MinInd])
                {
                    MinInd = i;
                }
            }
            NearestNode = NearNodeNum[MinInd];
            MinDistance = MinDist[MinInd];  
        }

        MPI_Bcast(&NearestNode, 1, MPI_INT, 0, MPI_COMM_WORLD);      

        // Определение номера процесса,
        // содержащего вершину с минимальным расстоянием       
        PivotProcess = NearestNode / RowNum; 

        // Включение вершины в состав МОД  
        if(0 == PivotProcess && ProcRank == 0)
        {
            ParentNodeNum = NotInSceleton[NearestNode - 1]->ParentNodeNum;
        }

        if(0 != PivotProcess)
        {      
            // Вершина не первая в массиве строк,
            // отправленных процессу
            if(NearestNode % RowNum != 0)
            {
                if(ProcRank == PivotProcess)
                {           
                    MPI_Send(&(NotInSceleton[NearestNode - 1]->ParentNodeNum), 
                        1, MPI_INT, 0, 1, MPI_COMM_WORLD);            
                }
            }

            // Вершина первая в массиве строк,
            // отправленных процессу
            if(NearestNode % RowNum == 0)
            {
                if(ProcRank == PivotProcess - 1)
                {
                    // Отправка от предыдущего процесса последней вершины
                    MPI_Send(&(NotInSceleton[NearestNode - 1]->ParentNodeNum), 
                        1, MPI_INT, PivotProcess, 2, MPI_COMM_WORLD); 
                }
                if(ProcRank == PivotProcess)
                {
                    MPI_Recv(&TmpNodeNum, 1, MPI_INT, PivotProcess-1, 2, MPI_COMM_WORLD, &Status);
                    MPI_Send(&TmpNodeNum, 1, MPI_INT, 0, 1, MPI_COMM_WORLD); 
                }
            }
        }// if(0 != PivotProcess)

        if(0 == ProcRank)
        {
            if(0 != PivotProcess)
            {             
                MPI_Recv(&ParentNodeNum, 1, MPI_INT, PivotProcess, 1, MPI_COMM_WORLD, &Status);
            }         

            // Присваивание информации о вершине МОД
            pSceleton[NearestNode] = (TTreeNode*) malloc (sizeof(TTreeNode));
            pSceleton[NearestNode]->NodeNum = ParentNodeNum;
            pSceleton[NearestNode]->Distance = MinDistance;
            pSceleton[NearestNode]->pNext = NULL;

            Parent = ParentNodeNum;
            if (pSceleton[Parent] != NULL)
            {
                tmp = (TTreeNode*) malloc (sizeof(TTreeNode));
                tmp->pNext = pSceleton[Parent]->pNext;
                tmp->Distance = MinDistance;
                tmp->NodeNum = NearestNode;
                (TTreeNode*)(pSceleton[Parent]->pNext) = tmp;
            }
            else
            {
                pSceleton[Parent] = (TTreeNode*) malloc(sizeof(TTreeNode));
                pSceleton[Parent]->pNext = NULL;
                pSceleton[Parent]->Distance = MinDistance;
                pSceleton[Parent]->NodeNum = NearestNode;
            }

            LastAdded = NearestNode;         
        }// if(0 == ProcRank)

        // Рассылка индекса последней добавленной вершины 
        // по всем процессам
        MPI_Bcast(&LastAdded, 1, MPI_INT, 0, MPI_COMM_WORLD);       
        
        // Освобождение памяти
        free(NotInSceleton[NearestNode - 1]);
        NotInSceleton[NearestNode - 1] = NULL;
       
    }// for ( Iter = 1; Iter < Size; ++Iter )

    free(NotInSceleton);

    return Res;
}
