﻿/*
//  Реализация алгоритма Флойда, 
//  поиска минимальных расстояний 
//  между всеми парами вершин графа с
//  использованием технологии OpenMP.
//
//  API
//  int OMP_FloydCalculation(double* pMatrix, const int Size);
//
//  В ходе выполнения алгоритма матрица смежности A изменяется,
//  после завершения вычислений в матрице A будет храниться требуемый результат - 
//  длины минимальных путей для каждой пары вершин исходного графа.
//
//  Организация параллельных вычислений состоит в одновременном выполнении
//  нескольких операций обновления значений матрицы смежности. Т.о. в качестве
//  базовой подзадачи используется операция обновления элементов матрицы смежности
//  (для указания подзадач используются индексы обновляемых в подзадачах элементов)
//
//  Input:
//      + pMatrix - указатель на вектор памяти элеметов типа double,
//	                матрица смежности графа
//	    + Size    - переменная типа int, размерность матрицы смежности
//  Output:
//      + pMatrix - указатель на вектор памяти элеметов типа double,
//	                изменённая матрица смежности, в которой будут храниться
//				    длины минимальных путей для каждой пары вершин исходного графа
//  Result:
//      + функция возвращает значение типа int, код заверщения функции
//
//
*/
int OMP_FloydCalculation(double* pMatrix, const int Size)
{
    int Res = PARLIB_SUCCESS;
    int i, j, k;
    double t1, t2;

    assert(Size > 0);
    assert(pMatrix);

    // В ходе выполнения алгоритма матрица смежности изменяется,
    // после завершения вычислений в матрице будет храниться требуемый результат - 
    // длины минимальных путей для каждой пары вершин исходного графа
    for( k = 0; k < Size; ++k )
    {
#pragma omp parallel for private (t1, t2, j)
        for( i = 0; i < Size; ++i )
        {
            for( j = 0; j < Size; ++j )
            {
                if((pMatrix[i * Size + k] != -1) &&
                    (pMatrix[k * Size + j] != -1)) 
                {
                        t1 = pMatrix[i * Size + j];
                        t2 = pMatrix[i * Size + k] + pMatrix[k * Size + j];
                        pMatrix[i * Size + j] = Min(t1, t2);
                }
            }
        }
    }
    return Res;
}

