﻿/*
//  Распределение данных по процессам для реализации решения
//  систем линейных однородных уравнений методом Гаусса.
//
//  API
//  int MPI_GaussianDataDistribution(double* pMatrix, double* pProcRows, 
//                                   double* pVector, double* pProcVector,
//                                   int* pProcInd, int* pProcNum,
//                                   const int ProcNum, const int ProcRank,
//                                   const int Size, int* RowNum);
//
//  Между всеми процессами распределяются строки матрицы коэффициентов системы,
//  значения компонент вектора правой части системы.
//
//  Input:
//      + pMatrix     - указатель на вектор памяти элементов типа double,
//                      матрица коэффициентов системы
//      + pProcRows   - указатель на векор памяти элементов типа double,
//                      часть строк матрицы коэффициентов системы,
//                      посылаемая процессу
//      + pVector     - указатель на вектор памяти элементов типа double,
//                      вектор правой части системы
//      + pProcVector - указатель на векор памяти элементов типа double,
//                      часть вектора правой части системы,
//                      посылаемая процессу
//      + pProcInd    - указатель на вектор памяти элементов типа int,
//                      память под индексы первых элементов последовательности
//                      строк матрицы, отправляемые процессу
//      + pProcNum    - указатель на вектор памяти типа int,
//                      память под число строк матрицы системы, 
//                      отправляемые процессу
//      + ProcNum     - переменная типа int, число всех процессов
//      + ProcRank    - переменная типа int, номер процесса
//      + Size        - переменная типа int, размерность системы
//      + RowNum      - указатель типа int
//  Output:
//      + pProcInd    - указатель на вектор памяти элементов типа int,
//                      индексы первых элементов последовательности
//                      строк матрицы, отправляемых процессу
//      + pProcNum    - указатель на вектор памяти типа int,
//                      число строк матрицы системы, 
//                      отправляемых процессу
//      + RowNum      - указатель на переменную типа int, число строк матрицы,
//                      отправляемых процессу
//  Result:
//      + функция возвращает значение типа int, код завершения функции
//
//
*/
int MPI_GaussianDataDistribution(double* pMatrix, double* pProcRows, 
                                 double* pVector, double* pProcVector,
                                 int* pProcInd, int* pProcNum,
                                 const int ProcNum, const int ProcRank,
                                 const int Size, int* RowNum)
{
    int Res = PARLIB_SUCCESS;


    int *pSendNum;     // Число элементов, отсылаемых поцессу
    int *pSendInd;     // Индекс первого элемента, отсылаемого процессу
    int RestRows=Size; // Число строк, которые ещё не распределены
    int i;           

    // Выделение памяти
    pSendInd = (int*) malloc (ProcNum * sizeof(int)); 
    if(NULL == pSendInd)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }
    pSendNum = (int*) malloc (ProcNum * sizeof(int));
    if(NULL == pSendNum)
    {
        Res = PARLIB_ERROR_OUT_OF_MEM;
    }

    // Определение расположения строк матрицы,
    // отсылаемых конкретному процессу
    *RowNum = (Size/ProcNum);
    pSendNum[0] = (*RowNum) * Size;
    pSendInd[0] = 0;
    for ( i = 1; i < ProcNum; ++i ) 
    {
        RestRows -= *RowNum;
        *RowNum = RestRows/(ProcNum-i);
        pSendNum[i] = (*RowNum) * Size;
        pSendInd[i] = pSendInd[i-1]+pSendNum[i-1];
    }

    // Рассылка строк
    MPI_Scatterv(pMatrix, pSendNum, pSendInd, MPI_DOUBLE, pProcRows, 
        pSendNum[ProcRank], MPI_DOUBLE, 0, MPI_COMM_WORLD);

    // Определение расположения части вектора - правой части системы,
    // отсылаемой конкретному процессу
    RestRows = Size;
    pProcInd[0] = 0;
    pProcNum[0] = Size/ProcNum;
    for ( i = 1; i < ProcNum; ++i ) 
    {
        RestRows -= pProcNum[i-1];
        pProcNum[i] = RestRows/(ProcNum-i);
        pProcInd[i] = pProcInd[i-1]+pProcNum[i-1];
    }

    // Рассылка частей вектора - правой части системы
    MPI_Scatterv(pVector, pProcNum, pProcInd, MPI_DOUBLE, pProcVector, 
        pProcNum[ProcRank], MPI_DOUBLE, 0, MPI_COMM_WORLD);

    // Освобождение памяти
    free(pSendNum);
    free(pSendInd);
    return Res;
}

/*
//  Сбор частей вектора результата воедино на нулевом процессе.
//
//  API
//  int MPI_GaussianResultCollection(double* pProcResult, double* pResult,
//                                   int* pProcNum, int* pProcInd,
//                                   const int ProcRank);
//
//  Input:
//      + pProcResult - указатель на вектор памяти типа double, 
//                      часть вектора результата
//      + pResult     - указатель на вектор памяти элементов типа double,
//                      вектор результата
//      + pProcNum    - указатель на вектор памяти типа int,
//                      число строк матрицы системы, 
//                      отправляемых процессу
//      + pProcInd    - указатель на векор памяти элементов типа int,
//                      индексы первых элементов последовательности
//                      строк матрицы, отправляемых процессу
//      + ProcRank    - переменная типа int, номер процесса
//  Output:
//      + pResult     - указатель на вектор памяти элементов типа double,
//                      вектор результата
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int MPI_GaussianResultCollection(double* pProcResult, double* pResult,
                                 int* pProcNum, int* pProcInd,
                                 const int ProcRank)
{
    int Res = PARLIB_SUCCESS;

    assert(pProcResult);
    assert(pResult);
    assert(pProcNum);
    assert(pProcInd);
    assert(ProcRank > -1);

    // Сбор всего вектора-результата на каждом процессе
    MPI_Gatherv(pProcResult, pProcNum[ProcRank], MPI_DOUBLE, pResult, 
                pProcNum, pProcInd, MPI_DOUBLE, 0, MPI_COMM_WORLD);


    return Res;
}

/*
//  Реализация алгоритма вычитания ведущей строки из строк,
//  которые еще не использовались в качестве ведущих.
//  Используется при вычислении решения системы линейных уравнений
//  методом Гаусса
//
//  API
//  int MPI_GaussianEliminateColumns(double* pProcRows, double* pProcVector,
//                                   double* pPivotRow, int* pProcPivotIter,
//                                   const int Size, const int RowNum,
//                                   const int Iter);
//
//  Input:
//      + pProcRows   - указатель на вектор памяти элементов типа double,
//                      строки матрицы коэффициентов системы
//      + pProcVector - указатель на вектор памяти элементов типа double,
//                      часть компонент правой части системы линейных уравнений
//      + pPivotRow   - указатель на вектор памяти элементов типа int,
//                      номера ведущих строк
//      + pPivotIter  - указатель на вектор памяти элементов типа int,
//                      номера итераций прямого хода метода Гаусса,
//                      на которых строки использовались в качестве ведущих
//      + Size        - переменная типа int, размерность системы
//      + RowNum      - переменная типа int, число строк матрицы, посланных процессу 
//      + Iter        - переменная типа int, номер столбца, в котором надо
//                      занулить все значения, кроме элемента в ведущей строке       
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int MPI_GaussianEliminateColumns(double* pProcRows, double* pProcVector,
                                 double* pPivotRow, int* pProcPivotIter,
                                 const int Size, const int RowNum,
                                 const int Iter)
{
    int Res = PARLIB_SUCCESS;
    int i, j;
    double multiplier; 

    assert(pProcRows);
    assert(pProcVector);
    assert(pPivotRow);
    assert(pProcPivotIter);
    assert(Size > 0);
    assert(RowNum > -1);
    assert(Iter > -1);

    // Исключение значений из соответствующнго столбца
    for ( i = 0; i < RowNum; ++i ) 
    {
        if (pProcPivotIter[i] == -1) 
        {
            multiplier = pProcRows[i * Size + Iter] / pPivotRow[Iter];
            for ( j  = Iter; j < Size; ++j ) 
            {
                pProcRows[i * Size + j] -= pPivotRow[j] * multiplier;
            }
            pProcVector[i] -= pPivotRow[Size] * multiplier;
        }
    }
    return Res;
}

/*
//  Реализация алгоритма
//  прямого хода метода Гаусса.
//
//  API
//  int MPI_GaussianElimination(double* pProcRows, double* pProcVector,
//                              int* pProcPivotIter, int* pParallelPivotPos,
//                              int* pProcInd, const int ProcRank,
//                              const int Size, const int RowNum);
//
//  Прямой ход метода Гаусса состоит в последовательном исключении неизвестных
//  в уравнениях решаемой системы линейных уравнений. 
//  На i-ой итерации метода производится исключение неизвестной i для всех уравнений
//  с номерами k, больших i. 
//  Для этого из этих уравнений осуществляется вычитание строки i,
//  умноженной на константу (Aki/Aii) с тем, 
//  чтобы результирующий коэффициент при неизвестной Xi в строках оказался нулевым.
//
//  Input:
//      + pProcRows          - указатель на вектор памяти элементов типа double,
//                             строки матрицы коэффициентов системы
//      + pProcVector        - указатель на вектор памяти элементов типа double,
//                             часть компонент правой части системы линейных уравнений
//      + pProcPivotIter     - указатель на вектор памяти элементов типа int,
//                             номера итераций прямого хода метода Гаусса,
//                             на которых строки использовались в качестве ведущих
//      + pParallelPivotPos  - указатель на вектор памяти элементов типа int,
//                             номера ведущих строк
//      + pProcInd           - указатель на векор памяти элементов типа int,
//                             индексы первых элементов последовательности
//                             строк матрицы, отправляемых процессу
//      + ProcRank           - переменная типа int, номер процесса
//      + Size               - переменная типа int, 
//                             размерность системы линейных уравнений
//      + RowNum            - переменная типа int, число строк матрицы, 
//                            посланных процессу 
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int MPI_GaussianElimination(double* pProcRows, double* pProcVector,
                            int* pProcPivotIter, int* pParallelPivotPos,
                            int* pProcInd, const int ProcRank,
                            const int Size, const int RowNum)
{
    int Res = PARLIB_SUCCESS;

    double MaxValue;   // Значение ведущего элемента на процессе
    int    PivotPos;   // Позиция ведущей строки в массиве строк,
                       // выделенных процессу 
    int i, j;
    // Структура для выбора ведущей строки   
    struct { 
        double MaxValue; 
        int ProcRank; 
    } ProcPivot, Pivot;

    // pPivotRow используется для хранения ведущей строки
    // и соответствующего элемента вектора - правой части системы
    double* pPivotRow = (double*) malloc ((Size+1) * sizeof(double));
    assert(pPivotRow);

    assert(pProcRows);
    assert(pProcVector);
    assert(pProcPivotIter);
    assert(pParallelPivotPos);
    assert(pProcInd);
    assert(ProcRank > -1);
    assert(Size > 0);
    assert(RowNum > -1);

    // Исключение элементов в столбцах
    for ( i = 0; i < Size; ++i )
    {
        // Вычисление локальной ведущей строки
        MaxValue = 0;
        for ( j = 0; j < RowNum; ++j )
        {
            if ((pProcPivotIter[j] == -1) && 
                (MaxValue < fabs(pProcRows[j*Size+i])))
            {
                    MaxValue = fabs(pProcRows[j*Size+i]);
                    PivotPos = j;
            }
        }
        ProcPivot.MaxValue = MaxValue;
        ProcPivot.ProcRank = ProcRank;

        // Поиск ведущего процесса (с наибольшим значением MaxValue)
        MPI_Allreduce(&ProcPivot, &Pivot, 1, MPI_DOUBLE_INT, MPI_MAXLOC, 
                      MPI_COMM_WORLD);

        // Рассылка ведущей строки
        if ( ProcRank == Pivot.ProcRank )
        {
            pProcPivotIter[PivotPos]= i; //iteration number
            pParallelPivotPos[i]= pProcInd[ProcRank] + PivotPos;
        }
        MPI_Bcast(&pParallelPivotPos[i], 1, MPI_INT, Pivot.ProcRank,
                  MPI_COMM_WORLD); 

        if ( ProcRank == Pivot.ProcRank )
        {
            // Копирование ведущей строки
            for ( j = 0; j < Size; ++j )
            {
                pPivotRow[j] = pProcRows[PivotPos*Size + j];
            }
            pPivotRow[Size] = pProcVector[PivotPos];
        }
        MPI_Bcast(pPivotRow, Size+1, MPI_DOUBLE, Pivot.ProcRank,
                  MPI_COMM_WORLD);

        MPI_GaussianEliminateColumns(pProcRows, pProcVector, pPivotRow, pProcPivotIter,
                                     Size, RowNum, i);
    }// for ( i = 0; i < Size; ++i )

    return Res;
}

/*
//  Реализация алгоритма
//  вычисления номера процесса, содержащего ведущую строку,
//  при выполнении обратной подстановки.
//  
//  API
//  int MPI_GaussianFindBackPivotRow(const int RowIndex, const int Size, 
//                                   const int ProcNum, int* pProcInd,
//                                   int* IterProcRank, int* IterPivotPos);
//
//  Input:
//      + RowIndex     - переменная типа int, номер ведущей строки
//      + Size         - переменная типа int, размерность системы                
//      + ProcNum      - переменная типа int, число процессов
//      + pProcInd     - указатель на векор памяти элементов типа int,
//                       индексы первых элементов последовательности
//                       строк матрицы, отправляемых процессу
//      + IterProcRank - указатель на переменную типа int
//      + IterPivotPos - указатель на переменную типа int
//  Output:
//      + IterProcRank - указатель на переменную типа int, номер процесса,
//                       содержащего ведущую строку
//      + IterPivotPos - указатель на переменную типа int,
//                       номер ведущего элемента в процессе
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int MPI_GaussianFindBackPivotRow(const int RowIndex, const int Size, 
                                 const int ProcNum, int* pProcInd,
                                 int* IterProcRank, int* IterPivotPos)
{
    int Res = PARLIB_SUCCESS;
    int i;

    assert(pProcInd);
    assert(RowIndex > -1);
    assert(Size > 0);

    // Вычисление номера ведущей строки
    for ( i = 0; i < ProcNum-1; ++i )
    {
        if ((pProcInd[i]<=RowIndex) && (RowIndex<pProcInd[i+1]))
        {
            *IterProcRank = i;
        }
    }
    if (RowIndex >= pProcInd[ProcNum-1])
    {
        *IterProcRank = ProcNum-1;
    }
    *IterPivotPos = RowIndex - pProcInd[*IterProcRank];


    return Res;
}

/*
//  Реализация алгоритма
//  обратного хода метода Гаусса.
//
//  API
//  int MPI_GaussianBackSubstitution(double* pProcRows, double* pProcVector,
//                                   int* pParallelPivotPos, int* pProcPivotIter,
//                                   double* pProcResult, int* pProcInd, 
//                                   const int Size, const int RowNum, 
//                                   const int ProcNum, const int ProcRank);
//
//  После приведения матрицы коэффициентов к верхнему треугольному виду
//  во время прямого хода метода Гаусса,
//  становится возможным определение значений неизвестных.
//  Из последнего уравнения преобразованной системы 
//  может быть вычислено значение переменной Xn-1,
//  после этого из предпоследнего уравнения становится возможным 
//  определение переменной Xn-2 и т.д.
//
//  Input:
//      + pProcRows         - указатель на вектор памяти элементов типа double,
//                            строки матрицы коэффициентов системы
//      + pProcVector       - указатель на вектор памяти элементов типа double,
//                            часть компонент правой части системы линейных уравнений
//      + pParallelPivotPos - указатель на вектор памяти элементов типа int,
//                            номера ведущих строк
//      + pProcPivotIter    - указатель на вектор памяти элементов типа int,
//                            номера итераций прямого хода метода Гаусса,
//                            на которых строки использовались в качестве ведущих
//      + pProcResult       - указатель на вектор памяти элеменов типа double,
//                            часть вектора результата
//      + pProcInd          - указатель на векор памяти элементов типа int,
//                            индексы первых элементов последовательности
//      + Size              - переменная типа int, 
//                            размерность системы линейных уравнений
//      + RowNum            - переменная типа int, число строк, 
//                            отправленных процессу
//      + ProcNum           - переменная типа int, число всех процессов
//      + ProcRank          - переменная типа int, номер процесса
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int MPI_GaussianBackSubstitution(double* pProcRows, double* pProcVector,
                                 int* pParallelPivotPos, int* pProcPivotIter,
                                 double* pProcResult, int* pProcInd, 
                                 const int Size, const int RowNum, 
                                 const int ProcNum, const int ProcRank)
{
    int Res = PARLIB_SUCCESS;

    int IterProcRank;    // Номер процесса, содержащего текущую ведущюу строку
    int IterPivotPos;    // Позиция ведущей строки в массиве строк,
                         // выделенных процессу
    double IterResult;   // Посчитанное значение неизвестной
    double val;
    int i, j;

    assert(pProcRows);
    assert(pProcVector);
    assert(pParallelPivotPos);
    assert(pProcPivotIter);
    assert(pProcResult);
    assert(pProcInd);
    assert(Size > 0);
    assert(RowNum > -1);
    assert(ProcRank > -1);
    assert(ProcNum > 0);

    // Обратная подстановка
    for ( i = Size - 1; i >= 0; --i )
    {

        // Вычисление номера процесса, содержащего ведущую строку
        MPI_GaussianFindBackPivotRow(pParallelPivotPos[i], Size, ProcNum, pProcInd, 
                                     &IterProcRank, &IterPivotPos);

        // Вычисление неизвестных
        if (ProcRank == IterProcRank) 
        {
            IterResult =
                pProcVector[IterPivotPos] / pProcRows[IterPivotPos * Size + i];
            pProcResult[IterPivotPos] = IterResult;
        }
        // Рассылка значения неизвестной на все процессы
        MPI_Bcast(&IterResult, 1, MPI_DOUBLE, IterProcRank, MPI_COMM_WORLD);

        // Изменение значений ветора - правой части системы
        for ( j = 0; j < RowNum; ++j )
        { 
            if ( pProcPivotIter[j] < i )
            {
                val = pProcRows[j * Size + i] * IterResult;
                pProcVector[j] = pProcVector[j] - val;
            }
        }
    }// for ( i = Size - 1; i >= 0; --i )

    return Res;
}

/*
//  Реализация алгоритма вычисления решения
//  системы n линейных уравнений методом Гаусса.
//
//  API
//  int MPI_GaussianResultCalculation(double* pProcRows, double* pProcVector, 
//                                    int* pProcPivotIter, double* pProcResult,
//                                    int* pParallelPivotPos, int* pProcInd,
//                                    const int ProcRank, const int ProcNum,
//                                    const int RowNum, const int Size);
//
//  В матричном виде система может быть представлена A*x = b,
//  где A - вещественная матрица коэффициентов при неизвестных размера n*n,
//  а вектора b и x состоят из n элементов.
//  Под задачей решения системы линейных уравнений для заданных матрицы А и вектора b
//  понимается нахождение значения вектора неизвестных x,
//  при котором выполняются все уравнения системы.
//
//  Input:
//      + pProcRows         - указатель на вектор памяти элементов типа double,
//                            строки матрицы коэффициентов системы
//      + pProcVector       - указатель на вектор памяти элементов типа double,
//                            часть компонент правой части системы линейных уравнений
//      + pProcPivotIter    - указатель на вектор памяти элементов типа int,
//                            номера итераций прямого хода метода Гаусса,
//                            на которых строки использовались в качестве ведущих
//      + pProcResult       - указатель на вектор памяти элеменов типа double,
//                            часть вектора результата
//      + pParallelPivotPos - указатель на вектор памяти элементов типа int,
//                            номера ведущих строк
//      + pProcInd          - указатель на векор памяти элементов типа int,
//                            индексы первых элементов последовательности
//      + ProcRank          - переменная типа int, номер процесса
//      + ProcNum           - переменная типа int, число всех процессов
//      + RowNum            - переменная типа int, число строк, 
//                            отправленных процессу
//      + Size              - переменная типа int, 
//                            размерность системы линейных уравнений
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int MPI_GaussianResultCalculation(double* pProcRows, double* pProcVector, 
                                  int* pProcPivotIter, double* pProcResult,
                                  int* pParallelPivotPos, int* pProcInd,
                                  const int ProcRank, const int ProcNum,
                                  const int RowNum, const int Size)
{
    int Res = PARLIB_SUCCESS;

    assert(pProcRows);
    assert(pProcVector);
    assert(pParallelPivotPos);
    assert(pProcPivotIter);
    assert(pProcResult);
    assert(pProcInd);
    assert(Size > 0);
    assert(RowNum > -1);
    assert(ProcRank > -1);
    assert(ProcNum > 0);

    // Прямой ход метода Гаусса
    MPI_GaussianElimination(pProcRows, pProcVector, pProcPivotIter, pParallelPivotPos, 
                            pProcInd, ProcRank, Size, RowNum);
    // Обратный ход метода Гаусса
    MPI_GaussianBackSubstitution(pProcRows, pProcVector, pParallelPivotPos, pProcPivotIter, 
                                 pProcResult, pProcInd, Size,
                                 RowNum, ProcNum, ProcRank);

    return Res;
}

