﻿/*
//  Реализация алгоритма вычисления решения
//  системы n линейных уравнений методом Гаусса,
//  с использованием технологии OpenMP.
//
//  API
//  int OMP_GaussianCalculation(double* pMatrix, double* pVector,
//                              double* pResult, const int Schedule,
//                              const int Size);
//
//  В матричном виде система может быть представлена A*x = b,
//  где A - вещественная матрица коэффициентов при неизвестных размера n*n,
//  а вектора b и x состоят из n элементов.
//  Под задачей решения системы линейных уравнений для заданных матрицы А и вектора b
//  понимается нахождение значения вектора неизвестных x,
//  при котором выполняются все уравнения системы.
//
//  В качестве базовой подзадачи определим все вычисления,
//  связанные с обработкой одной строки матрицы A
//  и соответствующего элемента вектора b
//
//  Input:
//      + pMatrix  - указатель на вектор памяти элементов типа double,
//                   матрица коэффициентов при неизвестных,
//                   i-я строка начинается с Size*i индекса
//      + pVector  - указатель на вектор памяти элементов типа double,
//                   правая часть системы линейных уравнений
//      + Schedule - переменная типа int, способ распределения строк между
//                   потоками
//      + Size     - переменная типа int, размерность системы линейных уравнений
//  Output:
//      + pResult  - указатель на вектор памяти элементов типа double,
//                  решение системы линейных уравнений
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int OMP_GaussianCalculation(double* pMatrix, double* pVector, double* pResult,
                            const int Schedule, const int Size)
{

    int Res = PARLIB_SUCCESS;
    int* pPivotPos;  // Определяет номера строк матрицы, выбираемых в качестве ведущих,
                     // по итерациям прямого хода метода Гаусса.
    int* pPivotIter; // Номера итераций прямого хода метода Гаусса,
                     // на которых строки использовались в качестве ведущих.
                     // Строка i выбиралась ведущей на итерации pPivotIter[i]
    int i;

    pPivotPos  = (int*) malloc (Size * sizeof(int));
    assert(pPivotPos);
    pPivotIter = (int*) malloc (Size * sizeof(int));
    assert(pPivotIter);

    // -1 - строка не выбиралась ещё в качестве ведущей
    for( i = 0; i < Size; ++i ) 
    {
        pPivotIter[i] = -1;
    }
    for( i = 0; i < Size; ++i )
        pResult[i] = 0.0;

    // Последовательное исключение неизвестных (прямой ход метода Гаусса)
    OMP_GaussianElimination(pMatrix, pVector, pPivotPos, pPivotIter, Schedule, Size);
    // Определение значений неизвестных (обратный ход метода Гаусса)
    OMP_BackSubstitution(pMatrix, pVector, pResult, pPivotPos, Size);


    free(pPivotPos);
    free(pPivotIter);

    return Res;
}

/*
//  Реализация алгоритма прямого хода метода Гаусса,
//  с использованием технологии OpenMP.
//
//  API
//  int OMP_GaussianElimination(double* pMatrix, double* pVector,
//                              int* pPivotPos, int* pPivotIter,
//                              const int Schedule, const int Size);
//
//  Прямой ход метода Гаусса состоит в последовательном исключении неизвестных
//  в уравнениях решаемой системы линейных уравнений. 
//  На i-ой итерации метода производится исключение неизвестной i для всех уравнений
//  с номерами k, больших i. 
//  Для этого из этих уравнений осуществляется вычитание строки i,
//  умноженной на константу (Aki/Aii) с тем, 
//  чтобы результирующий коэффициент при неизвестной Xi в строках оказался нулевым.
//
//  Input:
//      + pMatrix    - указатель на вектор памяти элементов типа double,
//                     матрица коэффициентов при неизвестных,
//                     i-я строка начинается с Size*i индекса
//      + pVector    - указатель на вектор памяти элементов типа double,
//                     правая часть системы линейных уравнений
//      + pPivotPos  - указатель на вектор памяти элементов типа int,
//                     номера строк матрицы, выбираемых в качестве ведущих,
//                     по итерациям прямого хода метода Гаусса
//      + pPivotIter - указатель на вектор памяти элементов типа int,
//                     номера итераций прямого хода метода Гаусса,
//                     на которых строки использовались в качестве ведущих
//      + Schedule   - переменная типа int, способ распределения строк между
//                     потоками
//      + Size       - переменная типа int, размерность системы линейных уравнений
//  Output:
//      + pResult    - указатель на вектор памяти элементов типа double,
//                     решение системы линейных уравнений
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int OMP_GaussianElimination(double* pMatrix, double* pVector,
                            int* pPivotPos, int* pPivotIter,
                            const int Schedule, const int Size)
{
    int Res = PARLIB_SUCCESS;
    int Iter;     
    int PivotRow; 

    assert(pMatrix);
    assert(pVector);
    assert(Schedule >= 0);
    assert(Size > 0);

    for ( Iter = 0; Iter < Size; ++Iter ) 
    {
        // Найти ведущую строку
        PivotRow = OMP_FindPivotRow(pMatrix, pPivotIter, Iter, Size);
        pPivotPos[Iter] = PivotRow;
        pPivotIter[PivotRow] = Iter;
        // Исключить неизвестные в столбце Iter
        OMP_ColumnElimination(pMatrix, pVector, pPivotIter, PivotRow, Iter, Schedule, Size);
    }

    return Res;
}

/*
//  Реализация алгоритма поиска ведущей строки 
//  для выполнения метода Гаусса на данной итерации,
//  с использованием технологии OpenMP.
//
//  API
//  int OMP_FindPivotRow(double* pMatrix, int* pPivotIter,
//                       const int Iter, const int Size);
//  
//  В качестве ведущей строки выбирается та строка, 
//  которая ранее не использовалась в качестве ведущей,
//  с максимальным по абсолютному значению элементом, 
//  расположенным в столбце с номером Iter.  
//
//  Cначала все потоки параллельно определияют
//  "локальную" ведущую строку в рамках своей полосы строк матрицы,
//  сохраняя её в структуре TThreadPivotRow, 
//  затем осуществляется выбор "глобальной" ведущей строки
//  среди "локальных" ведущих строк потоков.
//
//  Input:
//      + pMatrix    - указатель на вектор памяти элементов типа double,
//                     матрица коэффициентов при неизвестных,
//                     i-я строка начинается с Size*i индекса
//      + pPivotIter - указатель на вектор памяти элементов типа int,
//                     номера итераций прямого хода метода Гаусса,
//                     на которых строки использовались в качестве ведущих
//      + Iter       - переменная типа int, номер текущего столбца
//                     для поиска максимального по модулю элемента
//      + Size       - переменная типа int, размерность системы линейных уравнений  
//  Result:
//      + функция возвращает значение типа int, номер ведущей строки
*/
int OMP_FindPivotRow(double* pMatrix, int* pPivotIter,
                     const int Iter, const int Size)
{
    int PivotRow = -1;    // Номер ведущей строки
    double MaxValue =  0; // Значение ведущего элемента
    int i;  

    assert(pMatrix);
    assert(Iter >= 0);
    assert(Size >= 0);

    // Выбор ведущей строки с максимальным элементом
#pragma omp parallel 
    {
        TThreadPivotRow ThreadPivotRow;
        ThreadPivotRow.MaxValue = 0;
        ThreadPivotRow.PivotRow = -1;
#pragma omp for private(i)
        for ( i = 0; i < Size; ++i ) 
        {
            if ((pPivotIter[i] == -1) && 
                (fabs(pMatrix[i * Size + Iter]) > ThreadPivotRow.MaxValue)) 
            {
                    ThreadPivotRow.PivotRow = i;
                    ThreadPivotRow.MaxValue = fabs(pMatrix[i * Size + Iter]);
            }
        }
#pragma omp critical
        {
            if (ThreadPivotRow.MaxValue > MaxValue)
            {
                MaxValue = ThreadPivotRow.MaxValue;
                PivotRow = ThreadPivotRow.PivotRow;
            }
        } // pragma omp critical

    } // pragma omp parallel
    return PivotRow;
}

/*
//  Реализация алгоритма обратного хода метода Гаусса,
//  с использованием технологии OpenMP.
//
//  API
//  int OMP_BackSubstitution (double* pMatrix, double* pVector, 
//                            double* pResult, int* pPivotPos, 
//                            const int Size);
//
//  После приведения матрицы коэффициентов к верхнему треугольному виду
//  во время прямого хода метода Гаусса,
//  становится возможным определение значений неизвестных.
//  Из последнего уравнения преобразованной системы 
//  может быть вычислено значение переменной Xn-1,
//  после этого из предпоследнего уравнения становится возможным 
//  определение переменной Xn-2 и т.д.
//
//  Обработка строк линейной системы на каждой итерации 
//  обратного хода происходит независимо, поэтому данные вычисления 
//  можно распределить между параллельными потоками.
//
//  Input:
//      + pMatrix    - указатель на вектор памяти элементов типа double,
//                     приведенная прямым ходом Гаусса матрица коэффициентов,
//                     i-я строка начинается с Size*i индекса
//      + pVector    - указатель на вектор памяти элементов типа double,
//                     приведенная прямым ходом Гаусса правая часть
//                     системы линейных уравнений
//      + pPivotPos  - указатель на вектор памяти элементов типа int,
//                     номера строк матрицы, выбираемых в качестве ведущих,
//                     по итерациям прямого хода метода Гаусса
//      + Size       - переменная типа int, размерность системы линейных уравнений
//  Output:
//      + pResult    - указатель на вектор памяти элементов типа double,
//                     решение системы линейных уравнений
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int OMP_BackSubstitution (double* pMatrix, double* pVector, double* pResult,
                          int* pPivotPos, const int Size) 
{
    int Res = PARLIB_SUCCESS;
    int i, j;
    int RowIndex, Row;

    assert(pMatrix);
    assert(pVector);
    assert(pResult);
    assert(Size > 0);

    // Обратная подстановка
    for ( i = Size - 1; i >= 0; --i ) {
        RowIndex = pPivotPos[i];
        pResult[i] = pVector[RowIndex] / pMatrix[Size * RowIndex + i];
        // Каждый поток исключает свой элемент в столбце
        // и вычитает из правой части системы
#pragma omp parallel for private (Row, j) 
        for ( j = 0; j < i; ++j ) 
        {
            Row = pPivotPos[j];
            pVector[Row] -= pMatrix[Row * Size + i]*pResult[i];
            pMatrix[Row * Size + i] = 0;
        }
    }

    return Res;
}


/*
//  Реализация алгоритма вычитания ведущей строки из строк матрицы,
//  которые еще не использовались в качестве ведущих, с использованием
//  технологии OpenMP.
//
//  API
//  int OMP_ColumnElimination(double* pMatrix, double* pVector, int* pPivotIter,
//                            const int PivotRow, const int Iter, 
//                            const int Schedule, const int Size);
//
//  Распределим преобразования строк между параллельными потоками
//  и выполним их параллельно.
//
//  Input:
//      + pMatrix    - указатель на вектор памяти элементов типа double,
//                     матрица коэффициентов, i-я строка начинается с Size*i индекса
//      + pVector    - указатель на вектор памяти элементов типа double,
//                     правая часть системы линейных уравнений
//      + pPivotIter - указатель на вектор памяти элементов типа int,
//                     номера итераций прямого хода метода Гаусса,
//                     на которых строки использовались в качестве ведущих
//      + PivotRow   - переменная типа int, номер ведущей строки
//      + Iter       - переменная типа int, номер столбца, в котором надо
//                     занулить все значения, кроме элемента в ведущей строке
//      + Schedule   - переменная типа int, способ распределения строк между
//                     потоками
//      + Size       - переменная типа int, размерность системы линейных уравнений	              
//  Result:
//      + функция возвращает значение типа int, код завершения функции
*/
int OMP_ColumnElimination(double* pMatrix, double* pVector, int* pPivotIter,
                          const int PivotRow, const int Iter, 
                          const int Schedule, const int Size)
{
    int Res = PARLIB_SUCCESS;
    int i, j;
    double PivotValue, PivotFactor; 

    assert(pMatrix);
    assert(pVector);
    assert(PivotRow >= 0);
    assert(Iter >= 0);
    assert(Schedule >= 0);
    assert(Size >= 0);

    PivotValue = pMatrix[PivotRow * Size + Iter];
    // Исключение неизвестных в стобце Iter 
    // с планировщиком нагрузки по умолчанию
    if( Schedule == 0 )
    {		
#pragma omp parallel for private(PivotFactor, i, j) 
        for ( i = 0; i < Size; ++i ) 
        {
            if (pPivotIter[i] == -1) 
            {
                PivotFactor = pMatrix[i * Size + Iter] / PivotValue;
                for ( j = Iter; j < Size; ++j ) 
                {
                    pMatrix[i * Size + j] -= PivotFactor * pMatrix[PivotRow * Size + j];
                }
                pVector[i] -= PivotFactor * pVector[PivotRow];
            }
        }
    }
    // Исключение неизвестных в стобце Iter 
    // с динамическим планировщиком нагрузки
    if(Schedule==1)
    {
#pragma omp parallel for private(PivotFactor, i, j) schedule(dynamic, 1)
        for ( i = 0; i < Size; ++i ) 
        {
            if ( pPivotIter[i] == -1 ) 
            {
                PivotFactor = pMatrix[i * Size + Iter] / PivotValue;
                for ( j = Iter; j < Size; ++j ) 
                {
                    pMatrix[i * Size + j] -= PivotFactor * pMatrix[PivotRow * Size + j];
                }
                pVector[i] -= PivotFactor * pVector[PivotRow];
            }
        }
    }
    return Res;
}
